#include <cstdlib>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <cstring>
#include <string.h>

#include "shape.h"
#include "shape_generation.h"
#include "tree_generation.h"



void change_color(shape &S){
  if(S.nodes.size()==2 && S.nodes[0]==0 && S.nodes[1]==1) S.c=color((unsigned char) 0);
  for(unsigned int k=0;k<S.shapeChild.size();k++){
    change_color(S.shapeChild[k]);
  }
}

using namespace std;

/// SHAPE DATABASE REPOSITORY NAMES
#define NumberOfDataBases 4
char DataBaseName[NumberOfDataBases][300]={
  "TriangulationDataBase/FormasAbstractas/",
  "TriangulationDataBase/Mattise/",
  "TriangulationDataBase/LaMano/",
  "TriangulationDataBase/Tarkett/",
};
char PaletteDataBaseName[200]="PaletteDataBase/";

void fprintf_demo_failure(char mes[300]){
  FILE * f;
  f = fopen ("demo_failure.txt", "w");
  if(f==NULL){
    printf("We can not open demo_failure.txt");
    return;
  }
  fprintf(f,"%s\n",mes);
  fclose(f);
  return;
}


int main(int argc, char *argv[])
{

//  char name[4][200]={"palette_1045.bmp","palette_1046.bmp","palette_1047.bmp","palette_1048.bmp"};
//
//  #pragma omp parallel for
//  for(int k=0;k<4;k++){
//    imageBMP Ibmp(name[k]);
//    Ibmp=Ibmp.sortHSV();
//    Ibmp.write(name[k]);
//  }
//  return 0;

  /// PARAMETERS ( UP TO argv[11] PARAMETERS ARE MANDATORY)
  /// argv[0] = exe
  /// argv[1] = input parameter file (choice from images or by upload) )
  /// argv[2] = palette image (choice from images or by upload (default value -1. If selected from images it return the index in the palette database)
  /// argv[3] = new shape selection (choice from images but not mandatory) (integer default value -1)
  /// argv[4] = resolution (choice using an slider from 0.1 to 4. (step 0.1)) (default value 1.))
  /// argv[5] = reshuffle shapes (choice using true/false button) (default value false)
  /// argv[6] = reassign colors (choice using true/false button) (default value false)
  /// argv[7] = sort shape drawing order (choice using true/false button) (default value true)
  /// argv[8] = data folder
  /// argv[9] = output parameter file name ("output.info")
  /// argv[10] = output image name 1 ("output.bmp")
  /// argv[11] = output image name 2 ("outputPer.bmp")
  /// argv[12] = background silhouette named "maskBackground.bmp" or first shape silhouette to be used (BMP files) (OPTIONAL)
  /// argv[13]  ... = shape silhouette to be used (BMP files) (OPTIONAL)

  /// PERIODICITY RENDER PROPERTY
  bool periodicity=false;

  /// MESSAGE STRING FOR "demo_failure.txt"
  char mes[300];

  /// WE CHECK THE NUMBER OF COMMAND LINE PARAMETERS
  if(argc<12){
    sprintf(mes,"argc=%d and it should be at least argc=12\n",argc);
    printf("%s\n",mes);
    fprintf_demo_failure(mes);
    return(1);
  }

  /// DATA FOLDER NAME
  char data[400];
  sprintf(data,"%s",argv[8]);

  /// WE READ SHAPE DATABASE
  vector < vector<shape> > ShapeDataBaseV;
  ShapeDataBaseV.resize(NumberOfDataBases);
  for(int k=0;k<NumberOfDataBases;k++){
    char name[400];
    sprintf(name,"%s/%s",data,DataBaseName[k]);
    //printf("%s\n",name);
    ReadShapeDataBase(ShapeDataBaseV[k],name);
    printf("Number of shapes in database = %d\n",(int) ShapeDataBaseV[k].size());
    if(ShapeDataBaseV[k].size()==0){
      sprintf(mes,"Problems reading shape database : %s \n",DataBaseName[k]);
      printf("%s\n",mes);
      fprintf_demo_failure(mes);
      return(2);
    }
  }

  /// WE MANAGE BACKGROUND MASK
  shape MaskBackground;
  int widthMaskBackground=0;
  int heigthMaskBackground=0;
  int IndexFirstSilhouette=12;
  if(argc>12){
    if(strcmp(argv[12],"maskBackground.bmp")==0){
       IndexFirstSilhouette=13;
       imageBMP iBMP(argv[12]);
       if(iBMP.size()==0){
        sprintf(mes,"Problems reading maskBackground.bmp\n");
        printf("%s\n",mes);
        fprintf_demo_failure(mes);
        return 8;
      }

      widthMaskBackground=iBMP.width();
      heigthMaskBackground=iBMP.height();
      printf("WE COMPUTE THE IMAGE INTENSITY\n");
      vector<float> If(iBMP.size());
      for(int k=0;k<(int) If.size();k++) If[k]=(float) .2126*iBMP[k].R_+.7152*iBMP[k].G_+.0722*iBMP[k].B_;

      printf("WE ZOOM, FILTER AND BINARIZE THE IMAGE INTENSITY\n");
      /*float ScaleRadius=*/zoom(If,widthMaskBackground,heigthMaskBackground,512.,1024.,2.,10.);

      printf("WE COMPUTE TRIANGULATION\n");
  //*this=triangulation(If,width,height,sigmaHarris,radius,harris_treshold_value,MinEdgeCos,MaxEdgeDistance);
      MaskBackground=shape(triangulation(If,widthMaskBackground,heigthMaskBackground,1.,3,0.,0.86,1.));
//      triangulation(const imageBMP &I,float sigmaHarris=1.,int radius=3,float harris_treshold_value=0.,float MinSqArea=512.,float MaxSqArea=1024.,
//                float sigmaBinarization=2.,float BinarizationSlope=10.,float MinEdgeCos=0.86,float MaxEdgeDistance=1.)

    }
  }
  /// WE MANAGE SILHOUETTES UPLOADED BY THE USER
  if(argc>IndexFirstSilhouette){
    ShapeDataBaseV.push_back(  vector<shape>(argc-IndexFirstSilhouette));
    #pragma omp parallel for
    for(int k=IndexFirstSilhouette;k<argc;k++){
      imageBMP iBMP(argv[k]);
      if(iBMP.size()==0){

        ShapeDataBaseV[ShapeDataBaseV.size()-1][k-IndexFirstSilhouette]=shape(triangulation());
        continue;
      }
      triangulation T(iBMP);
      if(T.t.size()==0){
        ShapeDataBaseV[ShapeDataBaseV.size()-1][k-IndexFirstSilhouette]=shape(triangulation());
        continue;
      }
      T.normalize();
      T.fix_triangle_orientation();
      T.update_n();
      T.rebuild();
      T.rebuild();
      T.rebuild();
      T.rebuild();
      /// WE SCALE THE TRIANGULATION IF THE SILHOUETTE NAME INCLUDES A ZOOM FACTOR
      double zoom=extract_zoom(argv[k]);
      if(zoom>0.0001){
        for(int k=0;k<(int) T.p.size();k++) T.p[k]=T.p[k]*zoom;
      }
      ShapeDataBaseV[ShapeDataBaseV.size()-1][k-IndexFirstSilhouette]=shape(T);
    }
    for(int k=0;k<(int) ShapeDataBaseV[ShapeDataBaseV.size()-1].size();k++){
      if(ShapeDataBaseV[ShapeDataBaseV.size()-1][k].T.t.size()==0){
        sprintf(mes,"Problems with the shape silhouette  %s\n",argv[k+IndexFirstSilhouette]);
        printf("%s\n",mes);
        fprintf_demo_failure(mes);
        //system("pause");
        return(3);
      }
    }
  }

  /// DEFINE SHAPE INDEXES FOR NEW SHAPES UPLOADED BY THE USER
  vector<int> Database_Index,Shape_Index;
  build_database_and_shape_index(Database_Index,Shape_Index,argc-IndexFirstSilhouette);

  ///WE READ PALETTE DATABASE
  string DefaultPalette(argv[2]);
  vector<imageBMP> PaletteDataBase;
  if(fopen (DefaultPalette.c_str(), "r")==NULL){
    char name[400];
    sprintf(name,"%s/%s",data,PaletteDataBaseName);
    ReadPaletteDataBase(PaletteDataBase,name);
    if(PaletteDataBase.size()==0){
      sprintf(mes,"Problems reading palette database : %s \n",PaletteDataBaseName);
      printf("%s\n",mes);
      fprintf_demo_failure(mes);
      return(4);
    }
  }
  else{
    PaletteDataBase.push_back(imageBMP(DefaultPalette.c_str()));
    if(PaletteDataBase[0].size()==0){
      sprintf(mes,"Problems reading palette %s\n",DefaultPalette.c_str());
      printf("%s\n",mes);
      fprintf_demo_failure(mes);
      return 5;
    }
  }

  /// WE READ THE PARAMETER FILE FOR SHAPE GENERATION
  float zoom=590.;
  int OrderToDrawShapes=0;
  float width_float=1.413793;
  float height_float=1.;//=1.;
  float scale=1.;//=1.;
  vector<float> border_thick;
  vector<color> background_color;
  vector<tree_node_generator_parameters> TngpV;
  shape S;

  /// WE READ INPUT PARAMETER FILE
  if(fopen (argv[1], "r")==NULL){
    printf("Creating a new shape\n");
    srand (time(NULL));
    rand();
    S=RandomShapeParameterConfiguration(width_float,height_float,border_thick,background_color,
                                      TngpV,PaletteDataBase,rand()%PaletteDataBase.size(),OrderToDrawShapes,scale,periodicity);
    if(argc>IndexFirstSilhouette){
      for(int k=0;k<(int) TngpV.size();k++){
        TngpV[k].shape_selection.clear();
        TngpV[k].shape_selection_scaling.clear();
        TngpV[k].shape_database=1004;
        TngpV[k].shape_index=-1;
      }
    }
    write_image_configuration_parameters(argv[9],S,width_float,height_float,border_thick,background_color,zoom,TngpV,OrderToDrawShapes,scale,
                                       DefaultPalette,Database_Index,Shape_Index);
    read_image_configuration_parameters(argv[9],S,width_float,height_float,border_thick,background_color,zoom,TngpV,OrderToDrawShapes,
                                        scale,DefaultPalette,Database_Index,Shape_Index);

  }
  else{
    string error=read_image_configuration_parameters(argv[1],S,width_float,height_float,border_thick,background_color,zoom,TngpV,OrderToDrawShapes,
                                        scale,DefaultPalette,Database_Index,Shape_Index);
    if(error.length()>1 ){
      sprintf(mes,"Input texture parameter file is incorrect. Problems with parameter %s\n",error.c_str());
      printf("%s\n",mes);
      fprintf_demo_failure(mes);
      return 6;
    }

//    zoom=26932;
    if(zoom>2160){ ///AAA
      sprintf(mes,"HEIGHT OF THE GENERATED IMAGE IN PIXELS = %lf  IS TOO LARGE. THE LARGEST VALUE FOR THE ONLINE FACILITY IS 2048 \n",zoom);
      printf("%s\n",mes);
      fprintf_demo_failure(mes);
      return 7;
    }


//    if(argc>IndexFirstSilhouette){
//      bool UserDefinedShapes=false;
//      for(int k=0;k<(int) TngpV.size();k++){
//        if(TngpV[k].shape_database==1004){
//          UserDefinedShapes=true;
//          break;
//        }
//      }
//      if(UserDefinedShapes==false){
//        for(int k=0;k<(int) TngpV.size();k++){
//          TngpV[k].shape_selection.clear();
//          TngpV[k].shape_selection_scaling.clear();
//
//          TngpV[k].shape_database=1004;
//          TngpV[k].shape_index=-1;
//        }
//      }
//    }
  }
  border_thick_correction(S,width_float,height_float,border_thick,TngpV);

  /// WE MANAGE THE CASE WHERE background_color.size()!=S.shapeChild.size()
  if(background_color.size()>S.shapeChild.size()){
    background_color.resize(S.shapeChild.size());
  }
  else if(background_color.size()<S.shapeChild.size()){
    int k0=background_color.size();
    srand (time(NULL));
    rand();
    int palette_index=TngpV.size()==0?0:TngpV[0].palette_index;
    palette_index=palette_index%PaletteDataBase.size();
    for(int k=k0;k<(int) S.shapeChild.size();k++){
      background_color.push_back(PaletteDataBase[palette_index][ami_randI(PaletteDataBase[palette_index].size())]);
    }
    for(int k=k0;k<(int) background_color.size() && k<(int) S.shapeChild.size();k++){ S.shapeChild[k].c=background_color[k]; }
  }


  /// WE MANAGE SELECTION OF A NEW PALETTE FROM THE DATABASE
  if(PaletteDataBase.size()==1){
    DefaultPalette=string(argv[2]);
  }

  if(fopen (DefaultPalette.c_str(), "r")==NULL){
    if(isdigit(argv[2][0])==true){
      int new_palette_index=atoi(argv[2]);
      if(new_palette_index>=0){
        new_palette_index=new_palette_index%PaletteDataBase.size();
        for(int k=0;k<(int) TngpV.size();k++){
          TngpV[k].palette_index=new_palette_index;
          if(TngpV[k].tesselation_palette_index>=0){
            TngpV[k].tesselation_palette_index=new_palette_index;
          }
        }
      }
      for(int k=0;k<(int) background_color.size() && k<(int) S.shapeChild.size();k++){
        background_color[k]=PaletteDataBase[new_palette_index][ami_randI(PaletteDataBase[new_palette_index].size())];
        S.shapeChild[k].c=background_color[k];
      }
    }
  }
  else{
//   for(int k=0;k<background_color.size();k++){
//        background_color[k]=PaletteDataBase[0][ami_randI(PaletteDataBase[0].size())];
//   }
//   for(int k=0;k<background_color.size() && k<S.shapeChild.size();k++){ S.shapeChild[k].c=background_color[k]; }

   for(int k=0;k<(int) TngpV.size();k++){
    if(TngpV[k].tesselation_palette_index>=0) TngpV[k].tesselation_palette_index=0;
    TngpV[k].palette_index=0;
   }
  }

  /// WE MANAGE NEW SHAPE SELECTION PARAMETER
  int new_shape_selection=atoi(argv[3]);
  if(new_shape_selection>=0){
    srand (time(NULL));
    rand();
    for(int k=0;k<(int) TngpV.size();k++){
     TngpV[k].shape_selection.clear();
     TngpV[k].shape_selection_scaling.clear();
     TngpV[k].shape_database_and_index_definition(new_shape_selection);
    }
  }

  /// WE MANAGE NEW SHAPE DISTRIBUTION PARAMETER
  bool new_shape_distribution=atoi(argv[5]);
  if(new_shape_distribution==true){
    srand (time(NULL));
    rand();
    for(int k=0;k<(int) TngpV.size();k++){
     TngpV[k].random_seed=rand();
    }
  }

  /// WE MANAGE NEW COLOR DISTRIBUTION PARAMETER
  bool new_color_distribution=atoi(argv[6]);
  if(new_color_distribution==true){
    srand (time(NULL));
    rand();
    int palette_index=TngpV.size()==0?0:TngpV[0].palette_index;
    palette_index=palette_index%PaletteDataBase.size();
    for(int k=0;k<(int) background_color.size();k++){
        background_color[k]=PaletteDataBase[palette_index][ami_randI(PaletteDataBase[palette_index].size())];
    }
    for(int k=0;k<(int) background_color.size() && k<(int) S.shapeChild.size();k++){ S.shapeChild[k].c=background_color[k]; }

    for(int k=0;k<(int) TngpV.size();k++){
     TngpV[k].random_seed_color=rand();
    }
    S.light_spot=ami_rand<0.5?0.:0.2+0.6*ami_rand;
    if(ami_rand<0.5) S.light_spot=-2.*S.light_spot;
    if(periodicity==true) S.light_spot=0;
    for(int k=0;k<(int) S.shapeChild.size();k++) S.shapeChild[k].light_spot=S.light_spot;
    point2d pl(ami_rand*width_float,ami_rand*height_float);
    S.light_spot_center=pl;
  }

  /// WE MANAGE SORTING SHAPES WHEN DRAWING
  if(atoi(argv[7])==true){
     srand (time(NULL));
     rand();
    if (OrderToDrawShapes<0) OrderToDrawShapes=0;
    else OrderToDrawShapes=-1;
  }

  /// WE MANAGE RESOLUTION PARAMETER
  float resolution=atof(argv[4]);
  scale*=resolution;

  /// WE MANAGE IF THE USER HAVE UPLOADED NEW SHAPES FOR THE FIRST TIME
  if(argc>IndexFirstSilhouette){
    int Nnew=0;
    for(int k=0;k<(int) TngpV.size();k++){
      if(TngpV[k].shape_database==1004){
        Nnew=1;
        break;
      }
    }
    //printf("Nnew=%d\n",Nnew); system("pause");
    if(Nnew==0){
      for(int k=0;k<(int) TngpV.size();k++){
        TngpV[k].shape_selection.clear();
        TngpV[k].shape_selection_scaling.clear();
        TngpV[k].shape_database=1004;
        TngpV[k].shape_index=-1;
      }
    }
  }


  /// WE MANAGE BACKGROUND MASK
  if(heigthMaskBackground>0){
    S.print();
    double backGroundRatio=(double) widthMaskBackground/heigthMaskBackground;
    float desp=0.;
    double Sum_border_thick=0;
    for(int k=0;k<(int) border_thick.size();k++) Sum_border_thick+=border_thick[k]+desp;
    printf("Sum_border_thick=%lf\n",Sum_border_thick);
    double aux=1.-2*Sum_border_thick;
    width_float=2*Sum_border_thick+aux*backGroundRatio;
    S.extrema[1]=point2d(width_float,height_float);
    point2d pmin=point2d(Sum_border_thick,Sum_border_thick);
    point2d pmax=point2d(width_float,height_float)+point2d(-Sum_border_thick,-Sum_border_thick);
    S.shapeChild[0].T=triangulation(pmin,pmax);
    S.shapeChild[0].extrema_update();
    pmin=point2d(0.,0.);
    pmax=point2d(width_float,height_float);
    for(int k=0,n=1;k<(int) border_thick.size();k++){
      if(border_thick[k]<=0.) continue;
      S.shapeChild[n].T=triangulation(border_thick[k],pmin,pmax);
      pmin=pmin+point2d(border_thick[k]+desp,border_thick[k]+desp);
      pmax=pmax-point2d(border_thick[k]+desp,border_thick[k]+desp);
      S.shapeChild[n++].extrema_update();
    }
    Sum_border_thick-=0.00415039062;
    aux=(aux+2.5*0.00415039062)/heigthMaskBackground;
    for(int k=0;k<(int) MaskBackground.T.p.size();k++){
      MaskBackground.T.p[k]=point2d(Sum_border_thick,Sum_border_thick)+MaskBackground.T.p[k]*aux;
    }
    S.shapeChild[0].T=MaskBackground.T;
    S.shapeChild[0].extrema_update();
    S.print();
    //system("pause");
  }

  /// WE GENERATE THE NEW SHAPE
//  write_image_configuration_parameters(argv[9],S,width_float,height_float,border_thick,background_color,zoom,TngpV,OrderToDrawShapes,scale,
//                                       DefaultPalette,Database_Index,Shape_Index);
//  read_image_configuration_parameters(argv[9],S,width_float,height_float,border_thick,background_color,zoom,TngpV,OrderToDrawShapes,
//                                        scale,DefaultPalette,Database_Index,Shape_Index);

//  vector<tree_node_generator_parameters> TngpV0=TngpV;
//  for(int k=0;k<(int) TngpV.size();k++){
//     //if(TngpV[k].nodes[0]!=0) continue;
//     TngpV[k].shape_diameter*=scale;
//     TngpV[k].border_thick*=scale;
//     TngpV[k].min_distance_contouring*=scale;
//     TngpV[k].max_distance_contouring*=scale;
//     TngpV[k].tree_node_generator(S,ShapeDataBaseV,PaletteDataBase);
//     //printf("k=%d TngpV.size()=%d\n",k,(int) TngpV.size()); system("pause");
//  }

  tree_generator(S,ShapeDataBaseV,PaletteDataBase,TngpV,scale);

  /// WE SAVE IN DISK NEW SHAPE PARAMETERS
  write_image_configuration_parameters(argv[9],S,width_float,height_float,border_thick,background_color,zoom,TngpV,OrderToDrawShapes,scale,
                                       DefaultPalette,Database_Index,Shape_Index);


//change_color(S); ///AAA
  /// WE DRAW THE NEW SHAPE IN AN IMAGE IN BMP FORMAT
  int width=zoom*width_float/height_float;
  int height=zoom;
  width=4*(width/4);
  height=4*(height/4);
  imageBMP I(width,height,color((unsigned char) 255));
  //if(TngpV.size()>0) srand(TngpV[0].palette_index);
  draw_shape(S,I,height_float/zoom,PaletteDataBase,OrderToDrawShapes,periodicity);
  /// WE APPLY A MEDIAN FILTER TO REDUCE PIXEL NOISE
  morphological_median2D(I,1);

  /// WE WRITE THE IMAGE
  I.write(argv[10]);
//return 0; ///AAA
  /// IMAGE WITH 45 SYMMETRY
  imageBMP I2=I.symmetry45();
  I2.write(argv[11]);

  return 0;

}
